; File Name   :	D:\user\pohlers\cd2\hobby\z9001\crt80\CRT80.COM
; Format      :	Binary file
; Base Address:	0000h Range: 0100h - 0680h Loaded length: 0580h

; neuer CONOUT-Treiber Christian Schiewe, 27.12.1989. s.a. MP-Artikel
; basiert auf CP/A-Code bioscrt.mac
; reass: V. Pohlers 04/2009, 01/2011

; Der Treiber wird im Shadow-RAM auf Adr. 7000h abgelegt.
; Im BIOS wird die CONOUT-Routine gepatcht.
; Das setzt ein CP/M mit Shadow-RAM und CCP auf Adr. CC00h voraus.

; Der Zeichensatz des Z9001 wird um deutsche Umlaute erweitert (ibm = mp_cga.bin):
; ibm ch.schiewe Umlaut epson
; 8E	0Bh	Ae	[
; 99	0Ch	Oe	\ .
; 9A	0Dh	Ue	]
; 84	1Bh	ae	{
; 94	1Ch	oe	|
; 81	1Dh	ue	}
; E1	1Eh	sz	~

; der Treiber realisiert ein SCP kompatibles Terminal
; - Realisierung der SCP1715-Steuerzeichen 1b 5e xx (Feldattribut) und 1b 5f xx (Spezialzeichen)
;   Andere ESC-Folgen fhren i.a. auf '^' (unzulaessig),
;   ausser Feldattribute invers und hell (Reaktion wie Steuerz. 84..87)
; - Umlaut-Umkodierung fuer deutschen Zeichensatz wie SCP1715-Steuerzeichen 0Eh und 0Fh
;   d.h. Umlaute werden nach Umschalten auf Deutsch auf den Zeichen {|}[\]~ ausgegeben.
;	00h	;nop
;	01h	;home
;	02h	;(auch 82h) cursor an
;	03h	;(auch 83h) cursor aus
;	04h	;(auch 84h) normal, nicht inv.	grn auf schwarz
;	05h	;(auch 85h) normal,       inv.	schwarz auf grn
;	06h	;(auch 86h) hell  , nicht inv.	gelb auf schwarz
;		(nur  87h) hell  ,       inv.
;	07h	;bell (akustisches zeichen)
;	08h	;cursor zurueck
;	0ah	;LF
;	0ch	;bs loeschen, cursor links oben
;	0dh	;CR
;	0eh	;umschalten auf 2. Zeichensatz
;	0fh	;umschalten auf 1. Zeichensatz
;	14h	;rest bs loeschen
;	15h	;cursor vorwaerts
;	16h	;rest der zeile loeschen
;	18h	;zeile loeschen, cursor an zeilanf.
;	1ah	;eine zeile hoch
;	1bh	;escape
;	1b zeile+80h spalte+80h		Kursor posit. SCP
;	1b zeile+00h spalte+00h		geht auch bei CP/A
;	1b 5e xx	Feldattribut setzen
;		mit xx= 01urggbh
;			  u		unterstreichen,
;			   r		invers (* wird untersttzt),
;			    gg		Zeichengenerator-Umschaltung,
;			      b		blinken,
;			       h	heller (* wird untersttzt)
;	1b 5f xx	Spezialzeichen setzen
;		mit xx= 01ssssbh
;			  ssss		Zeichencode (1100 Sonderfkt)
;			      b		blinken des Sonderzeichens
;			       h	heller des Sonderzeichens
;	1ch	;Farbbefehl
;    	1c 40h+Farbe 			Randfarbe
;    	1c 10h+Farbe 			Hintergrundfarbe
;    	1c 20h+Farbe 			Vordergrundfarbe
;		+80h				blinkend
; 		0 schwarz
; 		1 rot
; 		2 grn
; 		3 gelb
; 		4 blau
; 		5 purpur (violett)
; 		6 cyan (hellblau)
; 		7 weiss
;	7fh	;delete
;	87h	;(nicht 07h) hell  ,      inv.

;vp: 87h wird meiner Meinung nach nicht richtig behandelt, das muss mal ausprobiert werden (basic print chr$)

		cpu	z80

CCP		equ	0CC00h		; base of ccp
UBIOS		equ	CCP+1600h	; base of bios

rambank1        equ	04h  		; ShadowRAM deaktivieren
rambank2        equ	05h  		; ShadowRAM aktivieren

crt80sw		equ	0		; 1 - SW-Verion

; Erweiterungen vp. Alles 0 -> orig. Version Schiewe
uz80zhw		equ	1		; 1 - U.Zander 80-Zeichen-Hardware
p_calcccp	equ	1		; CCP-Anfang autom. ermitteln

		if uz80zhw = 0
; Hardwarenansteuerung sicher analog mp 06/90, aber andere Adressen:
CS1	equ	0B9h		; out B9h	CS 1
CS2	equ	0BBh		; out BBh	CS 2
		else
;;neue Hardwareports nach U. Zander ("CRT81", 06.02.2011)
CS1	equ	0BEh		; out BEh	CS 1
CS2	equ	0BFh		; out BFh	CS 2
		endif


; vorgegebene Bildschirm-Werte:

bsanf1	equ	8000h		;Anfang Bildschirmpuffer BAB1
bszl1	equ	24		;dito BAB1
bssp1	equ	80		;dito BAB1
bsend1	equ	bsanf1+bszl1*bssp1-1	;dito BAB1



		org 100h

;------------------------------------------------------------------------------
; Einbindung ins BIOS
;------------------------------------------------------------------------------

		if uz80zhw = 1
		out	0BBh, a		; 2. Zeichensatz
		out	0BDh, a		; 80-Zeichen-Modus
		endif
; Kopieren in Shadow-RAM
		out	(rambank2), a	; ShadowRAM aktivieren
		ld	hl, p_crto	; Treiberposition im COM-File
; Treiber in Shadow-RAM kopieren
		ld	de, crto	; Ziel-Adr.
		ld	bc, p_crtoe-p_crto
		ldir
		out	(rambank1), a	; ShadowRAM deaktivieren

; Patchen des BIOS-Sprungverteilers
		if p_calcccp
		ld	hl, (1)		; Adresse von WBOOT (BIOS CALL 1)
		ld	de, 9+1		; 3* 3 Byte+1; BIOS CALL 4 (CONOUT)
		add	hl, de
		ld	e, (hl)
		inc	hl
		ld	d, (hl)		; DE=Ad. von CONOUT-Prozedur im UBIOS
;;		ld	hl, (UBIOS + 4*3 +1)	; UBIOS Sprungvektortabelle: _conout-Treiber-Adresse
;;		ex	de, hl
		ld	hl, _conout	; Treiberpatch fr Upper Bios
		ld	bc, 0Ah
		ldir
		else
		ld	hl, (UBIOS + 4*3 +1)	; UBIOS Sprungvektortabelle: _conout-Treiber-Adresse
		ex	de, hl
		ld	hl, _conout	; Treiberpatch fr Upper Bios
		ld	bc, 0Ah
		ldir
		endif

; Startmeldung ausgeben
		out	(CS1),	a	; CS 1 aktiv
		ld	de, startmsg
		ld	c, 9
		call	5
		ret

startmsg:	db 	0Ch			; cls
		db	1Ch,27h			; Vordergrund weiss
		db	0Fh			; deutscher ZS
		db	1Ch,14h,"    80 Zeichentreiber f}r "	; Hintergrund blau
		db	1Ch,0A3h,"KC87"		; blinkend, Vordergrund gelb
		db	1Ch,27h," installiert "	; Vordergrund weiss
		db	1Ch,20h,1Ch,11h		; Vordergrund scharz, Hintergrund rot
		db	" von Chr. Schiewe am 27.12.1989 !    ",7,7,7
		if uz80zhw = 1
		db	1Ch,27h			; Vordergrund weiss
		db	1Ch,14h			; Hintergrund blau
		db	"    erweitert V. Pohlers 13.02.2011    "
		db	1Ch,20h			; Vordergrund scharz
		endif
		db	1Ch,17h,0Dh,0Ah," Farben : "	; Hintergrund weiss
		db	1Ch,11h,"          "	; Hintergrund rot
		db	1Ch,12h,"          "	; Hintergrund grn
		db	1Ch,13h,"          "	; Hintergrund gelb
		db	1Ch,14h,"          "	; Hintergrund blau
		db	1Ch,15h,"          "	; Hintergrund purpur
		db	1Ch,16h,"          "	; Hintergrund cyan
		db	1Ch,17h,"          "	; Hintergrund weiss
		db	1Ch,10h,1Ch,22h		; Hintergrund scharz, Vordergrund grn
		db	0Dh,0Ah,0Dh,0Ah		;
		db 	"^O -> deutscher Zeichensatz: { } | [ ] \\ ~",0Dh,0Ah
		db 	0Eh,"^N -> normaler  Zeichensatz: { } | [ ] \\ ~",0Dh,0Ah,"$"

;-------------------------------------------------------------------------------
; Routine wird nach (0E20Dh) kopiert, d.h. patchen des Upper Bios
; 80-Zeichen-Treiber (Upper Bios)
;-------------------------------------------------------------------------------
_conout:	di
		out	(rambank2), a	; ShadowRAM aktivieren
		call	crto	; Aufruf 80-Zeichen-conout-Treiber
		out	(rambank1), a	; ShadowRAM deaktivieren
		ei
		ret



;-------------------------------------------------------------------------------
		if p_calcccp = 0
		org	280h
		endif
p_crto
; 80-Zeichen-Treiber wird nach 7000h kopiert (Shadow-RAM)
		phase	7000h

;------------------------------------------------------------------------------
; CONOUT-Treiber f. CP/M-Shadow-BIOS
;------------------------------------------------------------------------------

crto:		push	bc		; merken zeichen in c
		ld	hl, (crtcps)	; Kursorposition
; loeschen Kursor, Zeichen unter Cursor restaurieren
		push	af
		if crt80sw
		ld	a, 0ECh
		else
		ld	a, 0E8h	   	; hier wird die zu nutzende BWS-Page gepatcht: ECxx oder E8xx
		endif
		ld	(co80page+1), a
		ld	a, (char1)	; Merkzelle Zeichen unter Cursor
		ld	(co80chr+1), a
		call	co80		; physische Zeichenausgabe
;
		pop	af
		call	crtoti		; Zeichen ausgeben
;				 	HL danach auf neuer Kursorpositon

; Setzen Kursor auf (HL)

		ld	(crtcps), hl	; cursorposition
conout1:	ld	a, 80h 		; hier wird gepatcht 0 bei ^C (Cursor aus)
dcurs		equ	$-1
		or	a
		jr	z, crtonc
; Cursor anzeigen
		if crt80sw
		ld	a, 0ECh
		else
		ld	a, 0E8h
		endif
		ld	(co80page+1), a
		if crt80sw
		ld	a, 7Fh		; Cursor voller Block
		else
		ld	a, 87h		; Farbattribut: blinkend, schwarz auf weiss
		endif
		ld	(co80chr+1), a
		call	co80		; physische Zeichenausgabe
		ld	(char1), a	; Merkzelle orig. Zeichen unter Cursor

crtonc:		pop	bc		;wiederherstellen c-reg
		ret

;------------------------------------------------------------------------------
; physische Zeichenausgabe
; in HL: aktuelle logische Position = 8000h+zeile*80+spalte
; Patchen von co80page+1 und co80chr+1
; out A: das vorherige Zeichen an Adresse (BWS+HL)
;------------------------------------------------------------------------------

co80:		push	hl
		srl	h		; HL/2
		rr	l
		jr	nc, co80page	; Sprung, wenn gerade Adresse
		di
		out	(CS2),	a	; ungerade: Auswahl-FF setzen
co80page:	ld	a, 0E8h	   	; hier wird die zu nutzende BWS-Page gepatcht: ECxx oder E8xx
		and	0BFh
		add	a, h		; BS-Adresse berechnen:	BS-Offset EC00h	(E800h)	addieren
		ld	h, a
		ld	a, (hl)
		push	af
co80chr:	ld	a, 20h ; ' '    ; hier wird das auszugebende Zeichen gepatcht
		ld	(hl), a		; Zeichen ausgeben
		out	(CS1),	a	; Auswahl-FF rcksetzen
		pop	af
		pop	hl
		ei
		ret

;========================================================
; interner Aufruf Bildschirmtreiber
; i HL	Zeiger in Bildschirmpuffer auf Kursorposition
; o HL	neuer Zeiger auf Kursorposition
;========================================================


crtoti:		ld	a, 0		; escape mode
desc		equ	$-1		; hier wird gepatcht
		or	a
		jp	nz, crto04	;ja
;--

farb:		ld	a, 0		; hier wird das	farbflag gepatcht (farb = 1C)
		or	a
		jr	z, crto02	; kein farb -> Zeichen ausgeben
;-------------------------------------
; Farbe
		ld	a, c
		bit	6, a		; > 40h?
		jr	z, farb1	; nein
; neue Randfarbe setzen
		and	7
		call	shl3		; Farbe 3 Bits nach links schieben
		ld	b, a		; Farbe merken
		in	a, (88h)	; PIO1AD
		and	011000111b	; Randfarbe rcksetzen (Bits 5-3)
		or	b		; neue Randfarbe
		out	(88h), a	; PIO1AD
;
farb1:		push	hl
		ld	hl, fattr	; Merkzelle Farbattribut Zeichen
		ld	a, (hl)
		bit	4, c		; Zeichen < 10h ?
		jr	z, farb2	; ja
; Vordergrundfarbe
		and	0F0h		; incl. Blink-Bit
		ld	b, a
		ld	a, c
		and	7
		or	b
farb2:		bit	5, c		; < 20h ?
		jr	z, farb3
; Hintergrundfarbe
		and	0Fh
		ld	b, a
		ld	a, c
		and	7
		call	shl4		; 4 Bits nach links schieben
		or	b
farb3:		res	7, a
		bit	3, c
		jr	z, farb4
		call	shl4		; 4 Bits nach links schieben
farb4:		bit	7, c
		jr	z, farb5
		set	7, a
farb5:		ld	(hl), a
		pop	hl
		xor	a

;-------------------------------------------------------------------------------
; 1C Farbbefehl folgt
;    40h+Farbe -> Randfarbe
;    10h+Farbe -> Hintergrundfarbe
;    20h+Farbe -> Vordergrundfarbe
;-------------------------------------------------------------------------------

chr1c:		ld	(farb+1), a	; 1C
		ret

; UP Bits shiften
shl4:		rlca			; 4 Bits nach links schieben
shl3:		rlca			; 3 Bits nach links schieben
		rlca
		rlca
		ret

;------------------------------------------------------------------------------
; logischer Treiber Zeichenausgabe Register C
;------------------------------------------------------------------------------

crto02:		ld	a, c
		cp	20h ; ' '
		jr	c, crto02a	; Zeichen < 20h	-> Sonderzeichen
		cp	80h
		jr	c, crtobs	; Zeichen < 80h	-> normales Zeichen
		cp	88h
		jr	nc, crtobs	; Zeichen > 87h normal ausgeben
		res	7, a		; sonst 80..87 -> 00..07 und als Sonderzeichen
crto02a:	jp	crto09		; Sonderzeichen
;
crtobs:		ld	e, a		; normales Zeichen
umlaut:		ld	a, 0FFh		; wenn FFh, dann deutscher Zeichensatz (^O)
					; wenn 00h, dann normaler Zeichensatz (^N)
					; (Feld wird gepatcht)
		or	a
		jr	z, out_e	; bei normaler Ausgabe

;sonst: Umlaute umkodieren
		if uz80zhw = 0
		ld	a, e
		cp	'['             ; Ae
		jr	nz, umlaut1
		ld	e, 0Bh
		jr	out_e
umlaut1:	cp	']'             ; Ue
		jr	nz, umlaut2
		ld	e, 0Dh
		jr	out_e
umlaut2:	cp	'{'             ; ae
		jr	nz, umlaut3
		ld	e, 1Bh
		jr	out_e
umlaut3:	cp	'}'             ; ue
		jr	nz, umlaut4
		ld	e, 1Dh
		jr	out_e
umlaut4:	cp	'\\'            ; Oe
		jr	nz, umlaut5
		ld	e, 0Ch
		jr	out_e
umlaut5:	cp	'|'             ; oe
		jr	nz, umlaut6
		ld	e, 1Ch
		jr	out_e
umlaut6:	cp	'~'             ; sz
		jr	nz, out_e
		ld	e, 1Eh
		else	; uz80zhw = 1
; 8E	0Bh	Ae	[
; 99	0Ch	Oe	\ .
; 9A	0Dh	Ue	]
; 84	1Bh	ae	{
; 94	1Ch	oe	|
; 81	1Dh	ue	}
; E1	1Eh	sz	~
		ld	a, e
		cp	'['             ; Ae
		jr	nz, umlaut1
		ld	e, 8Eh
		jr	out_e
umlaut1:	cp	']'             ; Ue
		jr	nz, umlaut2
		ld	e, 9Ah
		jr	out_e
umlaut2:	cp	'{'             ; ae
		jr	nz, umlaut3
		ld	e, 84h
		jr	out_e
umlaut3:	cp	'}'             ; ue
		jr	nz, umlaut4
		ld	e, 81h
		jr	out_e
umlaut4:	cp	'\\'            ; Oe
		jr	nz, umlaut5
		ld	e, 99h
		jr	out_e
umlaut5:	cp	'|'             ; oe
		jr	nz, umlaut6
		ld	e, 94h
		jr	out_e
umlaut6:	cp	'~'             ; sz
		jr	nz, out_e
		ld	e, 0E1h
		endif
; Zeichenausgabe Register E
out_e:		ld	a, 0ECh
		ld	(co80page+1), a
		ld	a, e
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
		ld	a, 0E8h
		ld	(co80page+1), a
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
		ex	de, hl		; de=aktuelle Cursor-Position
; evtl. Zeilenvorschub ntig?

;------------------------------------------------------------------------------
; 15 	CTRL/U cursor vorwaerts
;------------------------------------------------------------------------------

crtcur:		ld	hl, 10000h-bsend1	; muss gescrollt werden?
		add	hl, de		;bs voll?
		ex	de, hl
		inc	hl
		ret	nc		; nein
;
		ld	de, bsend1-bssp1+1
;ein Zeile hochscrollen
crto03:		push	de
;23 Zeilen scrollen
		ld	hl, 0EC00h+40
		ld	de, 0EC00h
		ld	bc, 23*40	; 23 Bildschirmzeilen
		push	bc
		push	de
		push	hl
		ldir			; Zeilenvorschub CS 1
		pop	hl
		pop	de
		pop	bc
		di
		out	(CS2), a
		push	bc
		ldir			; Zeilenvorschub CS 2
		ld	hl, 0E800h+40
		ld	de, 0E800h
		pop	bc
		push	bc
		push	de
		push	hl
		ldir			; Zeilenvorschub CS 2 Farbspeicher
		pop	hl
		pop	de
		pop	bc
		out	(CS1), a
		ei
		ldir			; Zeilenvorschub CS 1 Farbspeicher
; 24. Zeile leeren
		ld	hl, 0EFC0h
		ld	(hl), ' '
		ld	de, 0EFC0h-1
		ld	bc, 40
		push	bc
		push	de
		push	hl
		lddr			; letzte Zeile leeren CS 1
		pop	hl
		pop	de
		pop	bc
		di
		out	(CS2), a
		ld	(hl), ' '
		push	bc
		lddr			; letzte Zeile leeren CS 2
		ld	hl, 0EBC0h
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(hl), a
		ld	de, 0EBC0h-1
		pop	bc
		push	bc
		push	de
		push	hl
		lddr			; letzte Zeile leeren CS 2 Farbspeicher
		pop	hl
		pop	de
		pop	bc
		out	(CS1), a
		ei
		ld	(hl), a
		lddr			; letzte Zeile leeren CS 1 Farbspeicher
		pop	hl
		ret


;------------------------------------------------------------------------------
; 05 	  	CTRL/E schwarz auf grn
;------------------------------------------------------------------------------

crto85:		ld	a, 2
		ld	(fattr), a	; Farbattribut Zeichen
		ret

;------------------------------------------------------------------------------
; 04            CTRL/D grn auf schwarz
;------------------------------------------------------------------------------

crto84:		ld	a, 20h
		ld	(fattr), a	; Farbattribut Zeichen
		ret

;------------------------------------------------------------------------------
; 0E		CTRL/N normaler Zeichensatz
;------------------------------------------------------------------------------

crt2zs:		xor	a
		ld	(umlaut+1), a
		ret

;------------------------------------------------------------------------------
; 0F		CTRL/O deutscher Zeichensatz
;------------------------------------------------------------------------------

crt1zs:		ld	a, 0FFh
		ld	(umlaut+1), a
		ret

;------------------------------------------------------------------------------
; 06		CTRL/F gelb auf schwarz
;------------------------------------------------------------------------------

crto86:		ld	a, 30h
		ld	(fattr), a	; Farbattribut Zeichen
		ret


;------------------------------------------------------------------------------
; 87		vp: eigentlich msste das "hell, inv." (schwarz auf gelb) sein. Ist so aber Bldsinn???
;------------------------------------------------------------------------------

crto87:		ld	a, 20h		; grn auf schwarz
		jr	$

;------------------------------------------------------------------------------
; Sonderzeichenbehandlung
;------------------------------------------------------------------------------

; Behandlung Sonderzeichen
;=========================

crto09:		ex	de, hl		; test sonderzeichen
		ld	hl, dszta	; Sonderzeichentabelle
		ld	bc, dsztal	; Lnge Sonderzeichentabelle
		cpir
		jr	z, crto11	; gefunden
;
		ld	c, '^'          ; nicht-Sonderz. anzeigen
		ex	de, hl		; HL restaurieren
		jp	crto02		; falls nicht gefunden

; Sonderzeichensprungtabelle
crto11:		ld	hl, dswe-1  	; Sonderzeichen-UP-Sprungtabellen-Ende
		or	a		; Cy=0
		sbc	hl, bc		; bc ^= gef. Position
		sbc	hl, bc		; 2x rckwrts
		ld	b, (hl)
		dec	hl
		ld	c, (hl)		; bc:=ansprungadresse
		push	bc		;
		ld	l, e		;
		ld	h, d		; bs adresse
		ld	a, ' '          ; fuer BS-loeschen
crtnop:		ret			; ansprung sonderzeichen

;------------------------------------------------------------------------------
; ESC
;	1b zeile+80h spalte+80h		Kursor posit. SCP
;	1b zeile+00h spalte+00h		geht auch bei CP/A
;	1b 5e xx	Feldattribut setzen
;		mit xx= 01urggbh
;			   r		invers,
;			       h	heller
;	1b 5f xx	Spezialzeichen setzen
;			-> wird immer zu '^'
;------------------------------------------------------------------------------

; Behandlung Zeichen innerhalb Escape-Folgen
;===========================================
crto04:		dec	a		; Eintritt A > 0 (also 2 oder 1)
		ld	(desc), a	; naechstes escape flag setzen
;
		ld	a, c		; Zeichen mit offset nach A
		res	7, c		; loeschen evtl. offset 80h
		jr	z, crto07	; letztes Zeichen ESC-Folge
; vorletztes Zeichen ESC-Folge
;-----------------------------
		sub	5Eh		; Feldattribut oder Sonderzeichen?
		ld	(crtffl), a 	; wenn ja, so 00 bzw. 01
		ret	z		; 1b 5e xx
		dec	a
		ret	z		; 1b 5f xx
;Positionierung Zeile
		ld	a, bszl1-1	;letzte zeile
		cp	c		;nach letzter zeile?
		jr	c, crto05	;ja
		ld	a, c		;sonst auf gewuenschte zeile
crto05:		ld	hl, bsanf1	;zeile adressieren
		or	a		; wenn 0
		ret	z
		ld	de, bssp1
		ld	b, a
crto06:		add	hl, de
		djnz	crto06		; zeile finden
		ret

; letztes Zeichen ESC-Folge
;--------------------------
crto07:		ld	b, 0FFh		;Wert wird gepatcht
crtffl	equ	$-1			;auf 00 bei "1b 5e xx"; auf 01 bei "1b 5f xx"
		djnz	crtnsp		;nicht 1b 5f xx
; Sonderzeichen
crtfa3:		ld	a, '^'		; Spezialzeichen nicht realisierbar
		jp	crtobs
;
crtnsp:		inc	b		;1b 5e xx?
		jr	nz, crtnfa	;nein
; Feldattribut
		bit	4, a		;invers?
		jr	z, crtfa1	;nein
		set	0, b		; B ist = 0
crtfa1:		bit	0, a		;hell?
		jr	z, crtfa2	;nein
		set	1, b
crtfa2:		and	00101110b	;wurden noch andere Funktionen verlangt?
		jr	nz, crtfa3	;ja, kann die Hardware nicht
		ld	a, b		; B = 0-normal, 1-invers, 2-hell, 3-hell+invers
		jp	crtobs		;sonst Steuerzeichen auswerten
;vp B msste sein: 4-normal, nicht inv., 5 - normal, inv., 6 hell,nicht inv., 87 hell, inv.
;vp und hier msste jp crto09 stehen
;
crtnfa:
; Positionierung Spalte
		ld	a, bssp1-1	;spalte setzen
		cp	c		;0 .. letzte spalte
		jr	nc, crto08	;ja
		ld	c, a		;sonst auf letzte spalte
crto08:		ld	b, 0
		add	hl, bc
		ret

; ...Realisierung Sonderzeichen...
;=================================

; 0D	  ENTER	  Cursor an den Anfang der Zeile
crtcr:		ld	bc,bssp1
		ld	hl,bsend1
crto12:		or	a
		sbc	hl,bc		;suchen,bis hl < (crtcps)
		push	hl
		sbc	hl,de
		pop	hl
		jr	nc,crto12	;zeilenanf. noch nicht gefunden
		inc	hl
		ret
;-----------------------------
; 0A	  v	  CTRL/J  Cursor runter
crtlf:		ld	hl,bsend1-bssp1
crtm14		equ	$-2
		sbc	hl,de		;bs ende (anfang letzte zeile)
		jp	c,crto03	;aufschieben
		ld	hl,bssp1
crtm15		equ	$-2
		add	hl,de		;naechste zeile
		ret
;-----------------------------
; crtdel		delete
crtdel:
; hide cursor
		ld	(co80chr+1), a
		ld	a, 0ECh
		ld	(co80page+1), a
		call	co80		; Zeichenausgabe
		ld	a, 0E8h
		ld	(co80page+1), a
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
;----------------------------
; 08	  <	  CTRL/H  Cursor nach links
crtcl:		ld	hl,bsanf1	;cursor zurueck
		sbc	hl,de
		ex	de,hl
		ret	z
		dec	hl
		ret

;------------------------------------------------------------------------------
; 0C		  CTRL/L  Bildschirm lschen
;------------------------------------------------------------------------------

crtbsl:		ld	hl, 0EC00h
		ld	(hl), a
		ld	de, 0EC00h+1
		ld	bc, 24*40
		push	bc
		push	de
		push	hl
		push	af
		ldir			; CS1
		pop	af
		pop	hl
		pop	de
		pop	bc
		di
		out	(CS2),	a
		ld	(hl), a
		push	bc
		ldir			; CS2
;
		ld	hl, 0E800h
		ld	de, 0E800h+1
		pop	bc
		push	bc
		push	de
		push	hl
		ld	a, (fattr)	; Farbattribut Zeichen
		push	af
		ld	(hl), a
		ldir			; CS2
		pop	af
		pop	hl
		pop	de
		pop	bc
		out	(CS1),	a
		ei
		ld	(hl), a
		ldir			; CS1
		ld	hl, bsanf1	; Pos (0,0)
		ret

;------------------------------------------------------------------------------
; 14 	CTRL/T
;------------------------------------------------------------------------------

crtbsr:		ld	hl, bsend1
		sbc	hl, de
		ex	de, hl
		push	af
		ld	(co80chr+1), a
		ld	a, 0ECh
		ld	(co80page+1), a
		call	co80		; Zeichenausgabe
		ld	a, 0E8h
		ld	(co80page+1), a
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
		pop	af
		ret	z
;
		push	hl
		ld	b, d
		ld	c, e
		ld	d, h
		ld	e, l
		inc	de
crtbsr_1:	ld	a, ' '
		ld	(co80chr+1), a
		ld	a, 0ECh
		ld	(co80page+1), a
		call	co80		; Zeichenausgabe
		ld	a, 0E8h
		ld	(co80page+1), a
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
		inc	hl
		dec	bc
		ld	a, b
		or	c
		jr	nz, crtbsr_1
		pop	hl
		ret

;----------------------------
; 01	  CTRL/A cursor home (0,0)
crthom:		ld	hl, bsanf1
		ret
;----------------------------
; 1A	  INS
crtcup:		ld	hl, bsanf1+bssp1-1  ;eine zeile hoch
		sbc	hl, de		;in oberster zeile?
		ex	de, hl
		ret	nc		;nicht ausfuehren
		ld	de, 0-bssp1
		add	hl, de		;zeilenlaenge abziehen
		ret
;-----------------------------
; 18	  CTRL/X
crtzl:		call	crtcr		;zeile loeschen
		ex	de, hl
; 16 	  CTRL/V
crtzlr:		push	de		;restzeile loeschen
		call	crtcr
		add	hl, bc
		or	a
		sbc	hl, de		;zeilenrest
		ex	de, hl
crto13:		ld	a, ' '
		ld	(co80chr+1), a
		ld	a, 0ECh
		ld	(co80page+1), a
		call	co80		; Zeichenausgabe
		ld	a, 0E8h
		ld	(co80page+1), a
		ld	a, (fattr)	; Farbattribut Zeichen
		ld	(co80chr+1), a
		call	co80		; Zeichenausgabe
		inc	hl
		dec	e
		jr	nz, crto13
		pop	hl
		ret
;-----------------------------
; 1B	  ESC
crtesc:		ld	a, 2		;escape flag setzen
		ld	(desc), a
		ret
;-----------------------------
; 02      Cursor ein
crtcrt:		ld	a, 80h		;cursor an
		jr	crto14
;-----------------------------
; 03	  Cursor aus
crtcof:		xor	a		;cursor aus
crto14:		ld	(dcurs), a
		ret

;------------------------------------------------------------------------------
; 07              CTRL/G  Ausgabe eines Summertones
;------------------------------------------------------------------------------

crtclk:		ld	bc, 40h
		ld	a, 00000111b
		out	(80h), a	; CTC0,	Kassette, akustischen Geber
		ld	a, 10000000b
		out	(80h), a
		call	crtclk_2		; Speaker ein
crtclk_1:	djnz	$
		dec	c
		jr	nz, crtclk_1
		ld	a, 00000011b	; CTC Reset
		out	(80h), a
; Speaker ein/aus
crtclk_2:	in	a, (88h)	; PIO1AD
		xor	10000000b	; , A7 : Freigabe des akustischen Gebers auf der Tastatur
		out	(88h), a
		ret

;------------------------------------------------------------------------------
; Sonderzeichentabelle
;------------------------------------------------------------------------------
;
; Tabelle der Sonderzeichen, SCP kompatibel!
;==========================================

dszta:
		db	00h	;nop
		db	01h	;home
		db	02h	;(auch 82h) cursor an
		db	03h	;(auch 83h) cursor aus
		db	04h	;(auch 84h) normal, nicht inv.
		db	05h	;(auch 85h) normal,       inv.
		db	06h	;(auch 86h) hell  , nicht inv.
;				 (nur  87h) hell  ,       inv.
		db	07h	;bell (akustisches zeichen)
		db	08h	;cursor zurueck
		db	0ah	;LF
		if uz80zhw = 1
		db	0bh	;Zeile hoch (Kompatibel zu Z9001)
		endif
		db	0ch	;bs loeschen, cursor links oben
		db	0dh	;CR
 		db	0eh	;umschalten auf 2. Zeichensatz
		db	0fh	;umschalten auf 1. Zeichensatz
 		db	14h	;rest bs loeschen
		db	15h	;cursor vorwaerts
		db	16h	;rest der zeile loeschen
 		db	18h	;zeile loeschen, cursor an zeilanf.
 		db	1ah	;eine zeile hoch
		db	1bh	;escape
;		1b zeile+80h spalte+80h		Kursor posit. SCP
;		1b zeile+00h spalte+00h		geht auch bei CP/A
;		1b 5e xx	Feldattribut setzen (SCP1715 fuer 8275-Contr.)
;			mit xx= 01urggbh
;				  u		unterstreichen,
;				   r		invers,
;				    gg		Zeichengenerator-Umschaltung,
;				      b		blinken,
;				       h	heller
;		1b 5f xx	Spezialzeichen setzen (SCP1715 fuer 8275)
;			mit xx= 01ssssbh
;				  ssss		Zeichencode (1100 Sonderfkt)
;				      b		blinken des Sonderzeichens
;				       h	heller des Sonderzeichens
		db	1ch	;Farbbefehl (s.o.)
		db	7fh	;delete
		db	87h	;(nicht 07h) hell  ,      inv.
dsztal	equ	$-dszta	;Tabellenlaenge

; Reaktionsroutinen fuer Sonderzeichen

szadr		dw crtnop		; 00		CLTR/@  nop
		dw crthom		; 01	 	CTRL/A	cursor home
		dw crtcrt		; 02/82  CLLN	CTRL/B  (auch 82h) cursor on
		dw crtcof		; 03/83  STOP	CTRL/C  (auch 83h) cursor off
		dw crto84		; 04/84 	CTRL/D  grn auf schwarz
		dw crto85		; 05/85  	CTRL/E  schwarz auf grn
		dw crto86		; 06/86  	CTRL/F  gelb auf schwarz
		dw crtclk		; 07 		CTRL/G  bell
		dw crtcl		; 08  <		CTRL/H  cursor left
		dw crtlf		; 0A  v		CTRL/J  cursor down
		if uz80zhw = 1
		dw crtcup		; 0B               	eine zeile hoch
		endif
		dw crtbsl		; 0C 		CTRL/L  clear screen
		dw crtcr		; 0D  ENTER	CTRL/M  carriage return
		dw crt2zs		; 0E 		CTRL/N  normaler Zeichensatz
		dw crt1zs		; 0F 		CTRL/O	deutscher Zeichensatz
		dw crtbsr		; 14 		CTRL/T  ab hier bildschirm lschen
		dw crtcur		; 15 		CTRL/U	cursor vorwaerts
		dw crtzlr		; 16 		CTRL/V  ab hier zeile lschen
		dw crtzl		; 18 		CTRL/X  zeile loeschen, cursor an zeilanf.
		dw crtcup		; 1A  INS               eine zeile hoch
		dw crtesc		; 1B  ESC               escape
		dw chr1c		; 1C  LIST		Farbbefehl (s. chr1c)
;
		dw crtdel		; 7F                    delete
		dw crto87		; 87 			(nicht 07h) hell  ,      inv.

dswe:	equ	$

;------------------------------------------------------------------------------
; Merkzellen
;------------------------------------------------------------------------------

crtcps:		dw bsanf1		; Merkzelle HL aktuelle logische Cursor-Position = 8000h+zeile*80+spalte
fattr:		db 20h			; Merkzelle Farbattribut Zeichen (grn auf schwarz)
char1:		db 20h			; Merkzelle Zeichen unter Cursor

		dephase
p_crtoe
		end
